import os
import time
import errno
import select
import socket
import logging
import functools

__all__ = ['TCP']

Timeout = socket.timeout
OPENING, OPEN, CLOSED = range(3)


BLOCK_SIZE = 8192


ERR_MAP = {
    errno.ECONNREFUSED: ConnectionRefusedError,
    errno.ECONNRESET: ConnectionResetError,
    errno.ECONNABORTED: ConnectionAbortedError,
    errno.EPIPE: BrokenPipeError,
    errno.EBADF: OSError,
}


def to_error(err):
    if err:
        return ERR_MAP.get(err, ConnectionError)(err, os.strerror(err))


def create_connection(host, port):
    sock = socket.socket()
    sock.setblocking(False)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    res = sock.connect_ex((host, port))
    if res not in {0, errno.EINPROGRESS}:
        raise to_error(res)
    return sock


def wait_open(sock, timeout=None):
    if timeout is None:
        _, w, _ = select.select((), (sock,), ())
    elif timeout >= 0:
        _, w, _ = select.select((), (sock,), (), timeout)
    else:
        raise Timeout("timeout trying to connect")
    if w:
        err = sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err:
            raise to_error(err)
    else:
        raise Timeout("timeout trying to connect")


def stream(sock, buffer_size=BLOCK_SIZE, timeout=None):
    readers = sock,
    while True:
        start = time.monotonic()
        r, _, _ = select.select(readers, (), (), timeout)
        end = time.monotonic()
        if timeout is not None:
            timeout -= start - end
        if (timeout is not None and timeout <= 0) or not r:
            raise Timeout("read timeout")
        data = sock.recv(buffer_size)
        if not data:
            break
        yield data


def check_open(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        state = self._state
        if state is CLOSED:
            raise to_error(errno.EBADF)
        elif state is OPENING:
            self.wait_open()
        return f(self, *args, **kwargs)
    return wrapper


def close_on_error(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        try:
            return f(self, *args, **kwargs)
        except BaseException:
            self.close()
            raise
    return wrapper


class RawTCP:

    def __init__(self, host, port, eol=b"\n", timeout=None):
        self.eol = eol
        self.timeout = timeout
        self._buffer = b""
        self._sock = None
        # create a non blocking socket
        self._state = OPENING
        self._connection_time = time.monotonic()
        self._sock = create_connection(host, port)

    def __del__(self):
        self.close()

    def wait_open(self):
        if self._state is OPEN:
            return
        elif self._state is CLOSED:
            raise OSError("would block forever")
        elif self._state is OPENING:
            timeout = self.timeout
            if timeout is not None:
                timeout -= time.monotonic() - self._connection_time
                timeout = max(0, timeout)
            wait_open(self._sock, timeout=timeout)
        self._state = OPEN

    @close_on_error
    def _write(self, data):
        for start in range(0, len(data), BLOCK_SIZE):
            _, w, _ = select.select((), (self._sock,), (), self.timeout)
            if not w:
                raise to_error(errno.EPIPE)
            self._sock.sendall(data[start: start + BLOCK_SIZE])

    @close_on_error
    def _read(self, n, timeout=None):
        if self._buffer:
            data, self._buffer = self._buffer, b""
            return data
        timeout = self.timeout if timeout is None else timeout
        r, _, _ = select.select((self._sock,), (), (), timeout)
        if r:
            data = self._sock.recv(BLOCK_SIZE)
            if not data:
                raise ConnectionError("remote end closed")
            return data
        else:
            raise Timeout("timeout reading from socket")

    @close_on_error
    def _readline(self, eol=None, timeout=None):
        eol = self.eol if eol is None else eol
        timeout = self.timeout if timeout is None else timeout
        data, eo, left = self._buffer.partition(eol)
        if eo:
            self._buffer = left
            return data + eo
        for data in stream(self._sock, timeout=timeout):
            self._buffer += data
            data, eo, left = self._buffer.partition(eol)
            if eo:
                self._buffer = left
                return data + eo
        else:
            raise ConnectionError("remote end closed")

    def state(self):
        return self._state

    def close(self):
        self._state = CLOSED
        self._buffer = b""
        if self._sock is not None:
            try:
                self._sock.shutdown(socket.SHUT_RDWR)
                self._sock.close()
            except OSError:
                pass
            finally:
                self._sock = None

    @check_open
    def write(self, data):
        self._write(data)

    @check_open
    def read(self, n, timeout=None):
        return self._read(n, timeout=timeout)

    @check_open
    def readline(self, eol=None, timeout=None):
        return self._readline(eol=eol, timeout=timeout)

    @check_open
    def write_read(self, data, n, timeout=None):
        self._write(data)
        return self._read(n, timeout=timeout)

    @check_open
    def write_readline(self, data, eol=None, timeout=None):
        self._write(data)
        return self._readline(eol=eol, timeout=timeout)


def ensure_connection(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        made_connection = self._ensure_connected()
        try:
            return f(self, *args, **kwargs)
        except Timeout:
            raise
        except OSError:
            self.close()
            if made_connection:
                raise
            self._ensure_connected()
            return f(self, *args, **kwargs)
    return wrapper


class TCP:

    def __init__(self, host, port, eol=b"\n", timeout=None):
        self.host = host
        self.port = port
        self.eol = eol
        self.timeout = timeout
        self.connection_counter = 0
        self._sock = None
        self._log = logging.getLogger("icepap.TCP({})".format(host))

    def connected(self):
        return self._sock is not None and self._sock.state() is OPEN

    def _ensure_connected(self):
        if self.connected():
            return False
        self._sock = RawTCP(
            self.host, self.port, eol=self.eol, timeout=self.timeout
        )
        self._sock.wait_open()
        self.connection_counter += 1
        self._log.debug("reconnecting #%d...", self.connection_counter)
        return True

    def connect(self):
        self._ensure_connected()

    def close(self):
        if self._sock is not None:
            self._sock.close()

    @ensure_connection
    def write(self, data):
        self._log.debug("write -> %r", data)
        self._sock.write(data)

    @ensure_connection
    def read(self, n, timeout=None):
        reply = self._sock.read(n, timeout=timeout)
        self._log.debug("read <- %r", reply)
        return reply

    @ensure_connection
    def write_read(self, data, n, timeout=None):
        self._log.debug("write_read -> %r", data)
        reply = self._sock.write_read(data, n, timeout=timeout)
        self._log.debug("write_read <- %r", reply)
        return reply

    @ensure_connection
    def write_readline(self, data, eol=None, timeout=None):
        self._log.debug("write_readline -> %r", data)
        reply = self._sock.write_readline(data, eol=eol, timeout=timeout)
        self._log.debug("write_readline <- %r", reply)
        return reply



# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# TODO: better use the __all__ = [...] structure
from .communication import IcePAPCommunication
from .controller import IcePAPController
from .fwversion import *
from .utils import *

# The version is updated automatically with bumpversion
# Do not update manually
version = '3.5.0'



# !/usr/bin/env python
# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import argparse
import logging
import logging.config
import time
import os
import sys
from .communication import IcePAPCommunication
from .__init__ import version


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        },
        'simple': {
            'format': '%(levelname)-8s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'file': {
            'level': 'INFO',
            'class': 'logging.FileHandler',
            'filename': '',
            'mode': 'w',
            'encoding': 'utf-8',
            'formatter': 'verbose'
        }
    },
    'loggers': {
        'Application': {
            'handlers': ['console', 'file'],
            'propagate': True,
            'level': 'INFO',
        },
        'icepap': {
            'handlers': ['file'],
            'level': 'INFO',
            'propagate': True,
        }
    }
}


def end(log, err_no=0):
    log_file = LOGGING_CONFIG['handlers']['file']['filename']
    log.info('Log saved on: {0}'.format(log_file))
    for h in log.handlers:
        h.flush()
    sys.exit(err_no)


def get_parser():
    desc = 'IcePAP scripts, base on ethernet communication\n'
    desc += 'Version: {}.\n'.format(version)
    epi = 'Documentation: https://alba-synchrotron.github.io/pyIcePAP-doc/\n'
    epi += 'Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain.'
    fmt = argparse.RawTextHelpFormatter
    parse = argparse.ArgumentParser(description=desc,
                                    formatter_class=fmt,
                                    epilog=epi)
    ver = '%(prog)s {0}'.format(version)
    parse.add_argument('--version', action='version', version=ver)
    subps = parse.add_subparsers(help='commands')

    # -------------------------------------------------------------------------
    #                           IcePAP commands
    # -------------------------------------------------------------------------
    # Send raw command
    send_cmd = subps.add_parser('send', help='Send IcePAP raw '
                                             'commands')

    send_cmd.set_defaults(which='send')
    send_cmd.add_argument('host', help='IcePAP Host')
    send_cmd.add_argument('-p', '--port', default=5000, help='IcePAP port')
    send_cmd.add_argument('-t', '--timeout', default=3, help='Socket timeout')
    send_cmd.add_argument('command', help='Raw command',
                          type=str)
    send_cmd.add_argument('-d', '--debug', action='store_true',
                          help='Activate log level DEBUG')
    # -------------------------------------------------------------------------
    return parse


def get_filename(host, command, filename='', log=False):
    value = time.strftime('%Y%m%d_%H%m%S')
    ext = ['cfg', 'log'][log]
    if host == '':
        host = 'icepap'
    new_filename = '{0}_{1}_{2}.{3}'.format(value, host, command, ext)
    if log or filename == '':
        filename = new_filename
    return os.path.abspath(filename)


def main():
    args = get_parser().parse_args()

    log_file = get_filename(args.host, args.which, log=True)
    LOGGING_CONFIG['handlers']['file']['filename'] = log_file
    logging.config.dictConfig(LOGGING_CONFIG)
    log = logging.getLogger('Application')

    # Send Command
    if args.which == 'send':
        ipap_com = IcePAPCommunication(host=args.host, port=args.port,
                                       timeout=args.timeout)
        log.info(ipap_com.send_cmd(args.command))


if __name__ == '__main__':
    main()



import time
import collections.abc

from .axis import IcePAPAxis
from .utils import State


def get_ctrl_item(f, axes, default=None):
    try:
        return f(axes)
    except Exception:
        pass
    values = []
    for axis in axes:
        try:
            value = f(axis)[0]
        except RuntimeError:
            value = default
        values.append(value)
    return values


def get_item(motors, name, default=None):
    values = []
    for motor in motors:
        try:
            value = getattr(motor, name)
        except RuntimeError:
            value = default
        values.append(value)
    return values


class Group:

    def __init__(self, motors):
        if isinstance(motors, IcePAPAxis):
            motors = [motors]
        ctrls = set(motor._ctrl for motor in motors)
        assert len(ctrls) == 1, 'motors must be from same controller'
        self._controller = ctrls.pop()
        self._motors = motors
        self._names = None

    @property
    def controller(self):
        return self._controller

    @property
    def motors(self):
        return self._motors

    @property
    def names(self):
        if self._names is None:
            self._names = self.get_names()
        return self._names

    @property
    def axes(self):
        return [motor.axis for motor in self.motors]

    def get_names(self):
        return get_item(self.motors, "name")

    def get_acctime(self):
        return get_ctrl_item(self.controller.get_acctime, self.axes)

    def get_velocity(self):
        return get_ctrl_item(self.controller.get_velocity, self.axes)

    def get_pos(self):
        return get_ctrl_item(self.controller.get_pos, self.axes)

    def get_fpos(self):
        return get_ctrl_item(self.controller.get_fpos, self.axes)

    def get_states(self):
        return get_ctrl_item(self.controller.get_states, self.axes, State(0))

    def is_moving(self):
        return any(state.is_moving() for state in self.get_states())

    def get_power(self):
        return get_ctrl_item(self.controller.get_power, self.axes)

    def start_stop(self):
        self._controller.stop(self.axes)

    def start_move(self, positions, **kwargs):
        args = [[mot.addr, pos] for mot, pos in zip(self._motors, positions)]
        self._controller.move(args, **kwargs)

    def start_rmove(self, positions, **kwargs):
        args = [[mot.addr, pos] for mot, pos in zip(self._motors, positions)]
        self._controller.rmove(args, **kwargs)

    def wait_stopped(self, timeout=None, interval=10e-3):
        """Helper loop to wait for group to finish moving"""

        start = time.time()
        while self.is_moving():
            time.sleep(interval)
            if timeout:
                elapsed = time.time() - start
                if elapsed > timeout:
                    return False
        return True


def group(*objs):
    if len(objs) == 1 and isinstance(objs[0], Group):
        return objs[0]
    motors = []
    for obj in objs:
        if isinstance(obj, IcePAPAxis):
            motors.append(obj)
        elif isinstance(obj, collections.abc.Sequence):
            motors.extend(group(*obj).motors)
        else:
            motors.extend(obj.motors)
    return Group(motors)



import os
import time
import signal
import logging
import contextlib

import icepap.group


@contextlib.contextmanager
def ensure_power(obj, on=True):
    """
    Power context manager. Entering context ensures the motor(s) have power
    (or an exception is thrown). Leaving the context leaves motor(s) has we
    found them. The context manager object is an icepap.Group.

    :param obj: a Group, IcePAPAxis or a sequence of IcePAPAxis
    :param on: if True, ensures power on all motors and when leaving the
               context restores power off on the motors  that were powered
               of before. If False the reverse behavior is applied

    Example::

        from icepap import IcePAPController, ensure_power
        ipap = IcePAPController("ipap.acme.com")
        m1, m2 = ipap[1], ipap[2]
        m1.power = False
        m2.power = True
        assert (m1.power, m2.power) == (False, True)
        with ensure_power((m1, m2)) as group:
            assert (m1.power, m2.power) == (True, True)
            group.start_move((1000, 2000))
            group.wait_move()
        assert (m1.power, m2.power) == (False, True)
    """
    g = icepap.group.group(obj)
    ctrl = g.controller
    powers = g.get_power()
    to_power = [addr for addr, power in zip(g.axes, powers) if power != on]
    if to_power:
        ctrl.set_power(to_power, on)
    try:
        yield g
    finally:
        if to_power:
            ctrl.set_power(to_power, not on)


def is_moving(states):
    return any(state.is_moving() for state in states)


def calc_deltas(p1, p2):
    return [abs(a - b) for a, b in zip(p1, p2)]


def gen_move(group, target):
    with ensure_power(group):
        group.start_move(target)
        for event in gen_motion(group):
            yield event


def gen_rmove(group, target):
    with ensure_power(group):
        group.start_rmove(target)
        for event in gen_motion(group):
            yield event


def gen_motion(group):
    while True:
        states, positions = group.get_states(), group.get_pos()
        yield states, positions
        if not is_moving(states):
            break


def gen_rate_limiter(generator, period=0.1):
    last_update = 0
    for event in generator:
        nap = period - (time.monotonic() - last_update)
        if nap > 0:
            time.sleep(nap)
        yield event
        last_update = time.monotonic()


def interrupt_myself():
    os.kill(os.getpid(), signal.SIGINT)



# #############################################################################
# Library provide by Detector and Electronic ESRF group.
# http://www.esrf.eu/Instrumentation/DetectorsAndElectronics
# #############################################################################

import struct
import numpy


class LibDeepDeepLogMock:
    DBG_DATA = 0

    def level(self):
        return self.DBG_DATA

    def trace(self, *args):
        print('DEEPLOG TRACE: %r' % args)


log = LibDeepDeepLogMock()

# -------------------------------------------------------------------------
#  Global definitions
#

PARAMETER = 'P'
POSITION = 'A'
SLOPE = 'S'

BYTE = 'BYTE'
WORD = 'WORD'
DWORD = 'DWORD'
LWORD = 'LWORD'
FLOAT = 'FLOAT'
DFLOAT = 'DFLOAT'
UBYTE = 'UBYTE'
UWORD = 'UWORD'
UDWORD = 'UDWORD'
ULWORD = 'ULWORD'
ADDRUNSET = 0xFF


# -------------------------------------------------------------------------
# Class definition
#

class vdata:
    """IcePAP data vector

    # object creation
    v = vdata()

    # appending data
    data = list(...)
    addr = 3
    type = vdata.PARAMETER  | POSITION | SLOPE
    v.append(data, addr, type)
    v.append(data, addr, type, format=vdata.DWORD)
    v.append(data, addr, type, incremental=True)
    """

    # Private members
    __signature = 0xCAFE
    __incremental = 0x8000

    __type_code = {
        PARAMETER: 0x1000,
        POSITION: 0x2000,
        SLOPE: 0x4000
    }

    __type_desc = {
        PARAMETER: 'PARAMETER',
        POSITION: 'POSITION',
        SLOPE: 'SLOPE'

    }

    __format_code = {
        BYTE: 0x00,
        WORD: 0x01,
        DWORD: 0x02,
        LWORD: 0x03,
        FLOAT: 0x04,
        DFLOAT: 0x05,
        UBYTE: 0x10,
        UWORD: 0x11,
        UDWORD: 0x12,
        ULWORD: 0x13
    }

    __format_dtype = {
        BYTE: 'b',
        WORD: 'h',
        DWORD: 'l',
        LWORD: 'q',
        FLOAT: 'f',
        DFLOAT: 'd',
        UBYTE: 'B',
        UWORD: 'H',
        UDWORD: 'L',
        ULWORD: 'Q'
    }

    __header_format = '<HBBLLBBHd'
    """
    from DSP source code:
    typedef struct {
     uint16_t    signature;   // Signature must be VDAT_SIGNATURE (0xCAFE)
     uint16_t    offset_vers; // High byte: data offset in dwords = 6  (0x0018)
                              // Low byte: version of the data vector format
                              // (0)
     uint32_t    v_size;      // full vector size in dwords
     uint32_t    n_values;    // number of values in the vector
     uint16_t    compr_dtype; // high byte: compression algorithm:
                              //            0=uncompressed, 1=lzapp
                              // low byte:  data type
     uint16_t    flags_addr;  // coding flags and board address
     extdfloat_t firstvalue;  // first vector value if incremental
    } vdatheader_t;
    """

    def __init__(self):
        """Object constructor"""
        self._bytearray = bytearray(0)

    def append(self, data, addr, type, format=DWORD, incremental=False):
        """Append data to current"""

        # Minimum checks
        if(addr < 0) or (addr > 255):
            raise ValueError("invalid address, must be [0:255]")

        if len(data) > 0xFFFF:
            raise ValueError("too many data values, max: 0xFFFF")

        if type not in self.__type_code:
            raise ValueError("invalid type specified")

        if format not in self.__format_code:
            raise ValueError("invalid format specified")

        dtype = self.__format_code[format]
        flags = self.__type_code[type] + addr

        dformat = '<' + str(len(data)) + self.__format_dtype[format]

        if incremental:
            firstval = data[0]
            data = [0 if i == 0 else data[i]-data[i-1]
                    for i in range(len(data))]
            flags |= self.__incremental
        else:
            firstval = 0

        if format not in [FLOAT, DFLOAT]:
            data = list(map(int, data))

        header_size = struct.calcsize(self.__header_format)
        full_size = header_size + struct.calcsize(dformat)
        if full_size % 4:
            full_size += 4 - (full_size % 4)

        # build the byte array
        bin_column = bytearray(full_size)
        struct.pack_into(
            self.__header_format,
            bin_column,
            0,
            self.__signature,  # vdata signature
            0,                 # Version = 0
            header_size // 4,   # Data offset in dwords
            full_size // 4,     # Full vector size in dwords
            len(data),         # number of values in the vector
            dtype,             # Data type
            0,                 # no compression
            flags,             # format + address
            firstval           # first data value for incremental coding
            )
        struct.pack_into(dformat, bin_column, int(header_size), *data)

        # append the byte array
        self._bytearray += bin_column

    def bin(self):
        """Return an IcePAP binary compatible block"""
        return numpy.array(self._bytearray, dtype=numpy.int8)

    def type_to_str(self, flags):
        """Returns a string describing an encoded vector type"""
        flags &= 0xff00
        for type in self.__type_code:
            if flags == self.__type_code[type]:
                return self.__type_desc[type]
        return 'unknown'

    def addr_to_str(self, flags):
        """Returns the destination address from an encoded vector header"""
        return flags & 0xff

    def loginfo(self):
        """Print out on log information about the data vector"""

        # brute dump of bytes
        # TODO: replace the test by a log.istrace()
        if log.level() >= log.DBG_DATA:
            log.trace("data vector contains:")
            n = self.bin()
            print(' '.join('0x{0:02x}'.format(x & 0xff) for x in n))

        # minimum check
        header_size = struct.calcsize(self.__header_format)
        if len(self._bytearray) < header_size:
            return

        # loop over all slices of data vector
        idx = 0
        cpt = 1
        while len(self._bytearray[idx:]) >= header_size:

            # extract header information
            (signature, version, header_sz, full_sz, data_len, data_type,
             compression, flags, firstval) = struct.unpack_from(
                self.__header_format, bytes(self._bytearray[idx:]))

            # minimum check
            if signature != self.__signature:
                raise ValueError('corrupted data vector, missing signature')

            # convert size from DWORD to bytes
            full_sz *= 4

            #
            log.trace("#%d: data vector type: %s" %
                      (cpt, self.type_to_str(flags)))
            log.trace("#%d: destination addr: %d" %
                      (cpt, self.addr_to_str(flags)))
            log.trace("#%d: number of data  : %d" %
                      (cpt, data_len))
            log.trace("#%d: data vector size: %dbytes" %
                      (cpt, full_sz))

            # jump to next slice
            idx += full_sz
            cpt += 1



# setup.py
import sys
from setuptools import setup
from setuptools import find_packages

# The version is updated automatically with bumpversion
# Do not update manually
__version = '3.5.0'

# windows installer:
# python setup.py bdist_wininst

# patch distutils if it can't cope with the "classifiers" or
# "download_url" keywords

# ipython profile magic commands implementation
package_list = ['*ipy']

setup_requirements = []

TESTING = any(x in sys.argv for x in ["test", "pytest"])
if TESTING:
    setup_requirements += ['pytest-runner']
test_requirements = ['pytest', 'pytest-cov']

SPHINX = any(x in sys.argv for x in ["build_sphinx"])
if SPHINX:
    setup_requirements += ['sphinx', 'sphinx-argparse', 'sphinx_rtd_theme']


requires = [
    'numpy',
    'click<8  ;python_version<"3.6"',
    'click>=7 ;python_version>="3.6"',
    'prompt_toolkit>=3 ;python_version>="3.6"',
    'beautifultable>=1 ;python_version>="3.6"'
]

setup(
    name="icepap",
    description="Python IcePAP Extension",
    version=__version,
    author="Guifre Cuni et al.",
    author_email="ctbeamlines@cells.es",
    url="https://github.com/ALBA-Synchrotron/pyIcePAP",
    packages=find_packages(),
    package_data={'': package_list},
    include_package_data=True,
    license="GPLv3",
    long_description="Python IcePAP Extension for Win32, Linux, BSD, Jython",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.5',
        'Topic :: Communications',
        'Topic :: Software Development :: Libraries',
    ],
    entry_points={
        'console_scripts': [
            'icepap = icepap.__main__:main',
            'icepapctl = icepap.cli:cli'
        ]
    },
    install_requires=requires,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    python_requires='>=3.5',
)



import shutil
import threading
import collections

import click
import beautifultable
from prompt_toolkit import print_formatted_text
from prompt_toolkit.styles import Style
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.shortcuts.progress_bar import ProgressBar
from prompt_toolkit.shortcuts.progress_bar.formatters import (
    Formatter, Label, Text, Percentage, Progress, Bar, TimeLeft
)

from .group import Group
from .controller import IcePAPController
from .tools import (
    ensure_power,
    interrupt_myself,
    is_moving,
    calc_deltas,
    gen_move,
    gen_rmove,
    gen_motion,
    gen_rate_limiter)

# -----------------------------------------------------------------------------
# Progress bar stuff


class Position(Formatter):

    moving = "<moving>{}</moving>"
    stopped = "<stopped>{}</stopped>"
    pos_format = ">{width}g"

    def format(self, progress_bar, progress, width):
        state = progress.state
        template = self.moving if state.is_moving() else self.stopped
        fmt = "{{:{}}}".format(self.pos_format.format(width=width)).format
        pos = fmt(progress.position)
        return HTML(template).format(pos)

    def get_width(self, progress_bar):
        fmt = "{{:{}}}".format(self.pos_format.format(width='')).format
        lengths = [len(fmt(c.initial_position)) for c in progress_bar.counters]
        lengths += [len(fmt(c.final_position)) for c in progress_bar.counters]
        # +2 to account for - sign and space
        return D.exact(max(lengths) + 2)


class InitialPosition(Formatter):

    pos_format = ">{width}g"

    def format(self, progress_bar, progress, width):
        fmt = "{{:{}}}".format(self.pos_format.format(width=width)).format
        return fmt(progress.initial_position)

    def get_width(self, progress_bar):
        fmt = "{{:{}}}".format(self.pos_format.format(width='')).format
        lengths = (len(fmt(c.initial_position)) for c in progress_bar.counters)
        return D.exact(max(lengths))


class TargetPosition(Formatter):

    pos_format = ">{width}g"

    def format(self, progress_bar, progress, width):
        fmt = "{{:{}}}".format(self.pos_format.format(width=width)).format
        return fmt(progress.final_position)

    def get_width(self, progress_bar):
        fmt = "{{:{}}}".format(self.pos_format.format(width='')).format
        lengths = (len(fmt(c.final_position)) for c in progress_bar.counters)
        return D.exact(max(lengths))


DEFAULT_FORMATTERS = [
    Label(),
    Text(" ["),
    InitialPosition(),
    Text(" => "),
    TargetPosition(),
    Text("] "),
    Position(),
    Text(" "),
    Percentage(),
    Text(" "),
    Bar(sym_a="=", sym_b=">", sym_c=" "),
    Text(" "),
    Progress(),
    Text(" "),
    Text("eta [", style="class:time-left"),
    TimeLeft(),
    Text("]", style="class:time-left"),
    Text(" "),
]

PLAIN_FORMATTERS = [
    Label(),
    Text(" ["),
    InitialPosition(),
    Text(" => "),
    TargetPosition(),
    Text("] "),
    Position(),
    Text(" "),
    Percentage(),
]

SIMPLE_FORMATTERS = [
    Label(),
    Text(": "),
    Position(),
]

DEFAULT_STYLE = Style.from_dict({
        "moving": "DeepSkyBlue bold",
        "stopped": "bold",
        "time-left": "DarkKhaki",
        "preparing": "DarkKhaki",
        "ctrlc": "orange",
        "error": "red",
        "done": "green",
})

DEFAULT_TOOLBAR = HTML(" Press [<b>x</b>] or [<b>Ctrl-C</b>] Stop.")


def create_default_key_bindings():
    kb = KeyBindings()

    @kb.add("x")
    def _(event):
        interrupt_myself()

    return kb


Motion = collections.namedtuple(
    "Motion",
    "group start_states start_positions initial_positions final_positions")


class MotionHandler:

    def __init__(self, on_ok=None, on_error=None, always=None):
        self.on_ok = on_ok or (lambda: None)
        self.on_error = on_error or (lambda err: None)
        self.always = always or (lambda: None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_value is None:
            self.on_ok()
        else:
            self.on_error(exc_value)
        self.always()


def MotionProgressBar(motion, bar_options=None):

    def prepare_counter(counter):
        def update(state, position):
            counter.state = state
            counter.position = position
            counter.items_completed = abs(position - counter.initial_position)

        def reset(initial, final):
            counter.total = abs(final - initial)
            counter.initial_position = initial
            counter.final_position = final

        counter.update = update
        counter.reset = reset

    def update(states, positions):
        for counter, state, pos in zip(prog_bar.counters, states, positions):
            counter.update(state, pos)
        prog_bar.invalidate()

    def reset(motion):
        args = (
            prog_bar.counters,
            motion.initial_positions,
            motion.final_positions)
        for counter, initial, final in zip(*args):
            counter.reset(initial, final)
        prog_bar.invalidate()

    bar_options = bar_options or {}
    bar_options.setdefault("title", HTML("<moving>Preparing...</moving>"))
    bar_options.setdefault("formatters", DEFAULT_FORMATTERS)
    bar_options.setdefault("style", DEFAULT_STYLE)
    bar_options.setdefault("key_bindings", create_default_key_bindings())
    bar_options.setdefault("bottom_toolbar", DEFAULT_TOOLBAR)
    prog_bar = ProgressBar(**bar_options)
    prog_bar.update = update
    prog_bar.reset = reset
    names = motion.group.names
    for name, state, start, initial, final in zip(names, *motion[1:]):
        counter = prog_bar(label=name, total=abs(final - initial))
        prepare_counter(counter)
        counter.reset(initial, final)
        counter.update(state, start)
    return prog_bar


def _move(group, final_positions, refresh_period=0.1, bar_options=None):
    start_states = group.get_states()
    start_positions = group.get_pos()
    motion = Motion(
        group, start_states, start_positions, start_positions, final_positions
    )

    def motion_loop(gen):
        for states, positions in gen_rate_limiter(gen, refresh_period):
            prog_bar.update(states, positions)

    def on_ok():
        nonlocal title
        title = HTML("<done>Finished!</done>")

    def on_error(error):
        nonlocal title
        group.start_stop()
        kb = isinstance(error, KeyboardInterrupt)
        if kb:
            msg = "<ctrlc>Stopping... </ctrlc>"
        else:
            msg = "<error>Motion error: {!r}<error>. ".format(error)
            msg += "Waiting for motors to stop..."
        title = HTML(msg)
        motion_loop(gen_motion(group))
        title = HTML(msg + "[<done>DONE</done>]")

    def always_end_with():
        prog_bar.update(group.get_states(), group.get_pos())

    title = HTML("<preparing>Preparing...</preparing>")
    bar_options = bar_options or {}
    bar_options.setdefault("title", lambda: title)
    prog_bar = MotionProgressBar(motion, bar_options=bar_options)
    power = ensure_power(group)
    handler = MotionHandler(on_ok, on_error, always_end_with)
    with prog_bar, power, handler:
        title = HTML("<moving>Moving...</moving>")
        motion_loop(gen_move(group, final_positions))


def _rmove(group, deltas, refresh_period=0.1, bar_options=None):
    start_states = group.get_states()
    start_positions = group.get_pos()
    final_positions = [p + d for p, d in zip(start_positions, deltas)]
    motion = Motion(
        group, start_states, start_positions, start_positions, final_positions
    )

    def motion_loop(gen):
        for states, positions in gen_rate_limiter(gen, refresh_period):
            prog_bar.update(states, positions)

    def on_ok():
        nonlocal title
        title = HTML("<done>Finished!</done>")

    def on_error(error):
        nonlocal title
        group.start_stop()
        kb = isinstance(error, KeyboardInterrupt)
        if kb:
            msg = "<ctrlc>Stopping... </ctrlc>"
        else:
            msg = "<error>Motion error: {!r}<error>. ".format(error)
            msg += "Waiting for motors to stop..."
        title = HTML(msg)
        motion_loop(gen_motion(group))
        title = HTML(msg + "[<done>DONE</done>]")

    def always_end_with():
        prog_bar.update(group.get_states(), group.get_pos())

    title = HTML("<preparing>Preparing...</preparing>")
    bar_options = bar_options or {}
    bar_options.setdefault("title", lambda: title)
    prog_bar = MotionProgressBar(motion, bar_options=bar_options)
    power = ensure_power(group)
    handler = MotionHandler(on_ok, on_error, always_end_with)
    with prog_bar, power, handler:
        title = HTML("<moving>Moving...</moving>")
        motion_loop(gen_rmove(group, deltas))


def _rmove_multiple(group, deltas, refresh_period=0.1, bar_options=None):
    start_states = group.get_states()
    start_positions = group.get_pos()
    final_positions = [p + d for p, d in zip(start_positions, deltas)]
    dir_deltas = deltas
    rev_deltas = [-delta for delta in deltas]
    motion = Motion(
        group, start_states, start_positions, start_positions, final_positions
    )

    def motion_loop(gen):
        for states, positions in gen_rate_limiter(gen, refresh_period):
            prog_bar.update(states, positions)

    def on_ok():
        nonlocal title
        title = HTML("<done>Finished!</done>")

    def on_error(error):
        nonlocal title
        group.start_stop()
        kb = isinstance(error, KeyboardInterrupt)
        if kb:
            msg = "<ctrlc>Stopping... </ctrlc>"
        else:
            msg = "<error>Motion error: {!r}</error>. ".format(error)
            msg += "Waiting for motors to stop..."
        title = HTML(msg)
        motion_loop(gen_motion(group))
        title = HTML(msg + "[<done>DONE</done>]")

    def always_end_with():
        prog_bar.update(group.get_states(), group.get_pos())

    title = HTML("<preparing>Preparing...</preparing>")
    bar_options = bar_options or {}
    evt = threading.Event()
    keys = create_default_key_bindings()
    action = None

    @keys.add('right')
    def _(event):
        nonlocal action
        action = "right"
        evt.set()

    @keys.add('left')
    def _(event):
        nonlocal action
        action = "left"
        evt.set()
    bar_options["key_bindings"] = keys
    bar_options["bottom_toolbar"] = HTML(
        " Press [<b>x</b>] or [<b>Ctrl-C</b>] Stop | "
        '[<a bg="deepskyblue">left arrow</a>] Move left | '
        '[<a bg="deepskyblue">right arrow</a>] Move right.')
    bar_options.setdefault("title", lambda: title)
    prog_bar = MotionProgressBar(motion, bar_options=bar_options)
    power = ensure_power(group)
    handler = MotionHandler(on_ok, on_error, always_end_with)
    with prog_bar, power, handler:
        while True:
            evt.wait()
            evt.clear()
            if action == "stop":
                break
            curr_deltas = dir_deltas if action == "right" else rev_deltas
            start_pos = group.get_pos()
            final_pos = [p + d for p, d in zip(start_pos, curr_deltas)]
            motion = Motion(
                group, start_states, start_pos, start_pos, final_pos
            )
            prog_bar.reset(motion)
            action = None
            title = HTML("<moving>Moving...</moving>")
            motion_loop(gen_rmove(group, curr_deltas))
            title = HTML("<done>Finished!</done>")

# -----------------------------------------------------------------------------
# Tables


def bool_text(data, false="NO", true="YES"):
    return true if data else false


def bool_text_color(data, text_false="NO", text_true="YES",
                    color_false="bright_red",
                    color_true="green"):
    color = color_true if data else color_false
    text = bool_text(data, text_false, text_true)
    text = click.style(text, fg=color)
    return text


def bool_text_color_inv(data, text_false="NO", text_true="YES"):
    return bool_text_color(data, text_false, text_true,
                           "green",
                           "bright_red")


def Table(**kwargs):
    style = kwargs.pop("style", beautifultable.Style.STYLE_BOX_ROUNDED)
    kwargs.setdefault("default_padding", 1)
    kwargs.setdefault("maxwidth", shutil.get_terminal_size().columns - 1)
    kwargs.setdefault(
        "default_alignment",
        beautifultable.Alignment.ALIGN_RIGHT)
    table = beautifultable.BeautifulTable(**kwargs)
    table.set_style(style)
    return table


def StateTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    table = Table(style=style)
    table.columns.header = (
        "Axis", "Name", "Pos.", "Ready", "Alive", "Pres.", "Enab.",
        "Power", "5V", "Lim-", "Lim+", "Warn")
    args = (
        group.motors,
        group.names,
        group.get_states(),
        group.get_fpos()
    )
    for motor, name, state, pos in zip(*args):
        row = (motor.axis, name, pos,
               bool_text_color(state.is_ready()),
               bool_text_color(state.is_alive()),
               bool_text_color(state.is_present()),
               bool_text_color(not state.is_disabled()),
               bool_text_color(state.is_poweron(), "OFF", "ON"),
               bool_text_color(state.is_5vpower()),
               bool_text_color_inv(state.is_limit_negative(), "OFF", "ON"),
               bool_text_color_inv(state.is_limit_positive(), "OFF", "ON"),
               bool_text_color_inv(state.is_warning()))
        table.rows.append(row)
    return table


def StatusTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    table = Table(style=style)
    table.columns.header = ('Axis', 'Name', 'Pos.', 'Ready', 'Vel.', 'Acc. T.')
    args = (group.motors, group.names, group.get_states(), group.get_fpos(),
            group.get_acctime(), group.get_velocity())
    for motor, name, state, pos, acctime, velocity in zip(*args):
        row = (motor.axis, name, pos,
               bool_text_color(state.is_ready()), velocity, acctime)
        table.rows.append(row)
    return table


def PositionTable(group, style=beautifultable.Style.STYLE_BOX_ROUNDED):
    ctrl, axes = group.controller, group.axes
    table = Table(style=style)
    header = "Axis", "AXIS", "MEASURE", "ENCIN", "INPOS", "ABSENC", "MOTOR"
    table.columns.header = header
    cols = [
        ctrl.get_pos(axes, register=register)
        for i, register in enumerate(header[1:])
    ]
    for row in zip(axes, *cols):
        table.rows.append(row)
    return table


# ------------------------------------------------------------------------------
# Command line interface


class ProgressBarFormats(click.Choice):

    formatters = dict(
        default=DEFAULT_FORMATTERS,
        plain=PLAIN_FORMATTERS,
        simple=SIMPLE_FORMATTERS
    )

    def __init__(self):
        super().__init__(self.formatters)

    def convert(self, value, param, ctx):
        result = super().convert(value, param, ctx)
        return self.formatters[result]


class TableStyles(click.Choice):

    name = "table_style"

    def __init__(self):
        styles = [s.name.lstrip("STYLE_").lower()
                  for s in beautifultable.Style]
        super().__init__(styles, case_sensitive=False)

    def convert(self, value, param, ctx):
        result = super().convert(value, param, ctx)
        return beautifultable.Style["STYLE_" + result.upper()]


class Axes(click.ParamType):

    name = "axes"

    def convert(self, value, param, ctx):
        pap = ctx.obj["icepap"]
        if value == "all":
            axes = pap.find_axes(only_alive=False)
        elif value == "alive":
            axes = pap.find_axes(only_alive=True)
        else:
            axes = []
            for axis in value.split(","):
                try:
                    axes.append(int(axis))
                except ValueError:
                    axes.append(axis.strip())
        return pap[axes]


def Racks(value):
    return [int(v) for v in value.split(",")]


def cli_move(group, positions, format=None, bottom_toolbar=True, title=True):
    bar_options = dict(formatters=format)
    if not bottom_toolbar:
        bar_options["bottom_toolbar"] = None
    if title is not True:
        bar_options["title"] = None if title is False else title
    _move(group, positions, bar_options=bar_options)


def cli_rmove(
        group,
        deltas,
        format=None,
        bottom_toolbar=True,
        title=True,
        multiple=False):
    bar_options = dict(formatters=format)
    if not bottom_toolbar:
        bar_options["bottom_toolbar"] = None
    if title is not True:
        bar_options["title"] = None if title is False else title
    if multiple:
        _rmove_multiple(group, deltas, bar_options=bar_options)
    else:
        _rmove(group, deltas, bar_options=bar_options)


opt_url = click.option(
    "-u", "--url", "icepap", type=IcePAPController.from_url, required=True,
    help="hardware url (ex: 'ipap.acme.org' or 'tcp://ipap13:5001')"
)

opt_pb_format = click.option(
    "--pb-format", type=ProgressBarFormats(),
    default="default", show_default=True, help="progress bar style"
)

opt_toolbar = click.option(
    "--bottom-toolbar/--no-bottom-toolbar", default=True, show_default=True,
    help="show/hide bottom toolbar"
)

opt_title = click.option(
    "--title/--no-title", default=True, show_default=True,
    help="show/hide title"
)

opt_table_style = click.option(
    "--table-style", type=TableStyles(), default="box_rounded",
    show_default=True, help="table style"
)

opt_axes = click.option(
    "--axes", "motors", type=Axes(), default="all", show_default=True,
    help="comma separated list of axes. Also supports 'all' and 'alive'"
)

opt_mandatory_axes = click.option(
    "--axes", "motors", type=Axes(), required=True, show_default=True,
    help="comma separated list of axes. Also supports 'all' and 'alive'"
)

opt_rack = click.option(
    "--rack", type=int, default=None, show_default=True,
    help="rack number. Default is None, meaning the whole system"
)

opt_racks = click.option(
    "--racks", type=Racks, required=True, show_default=True,
    help="comma separated list of racks. Also supports 'all' and 'alive'"
)


@click.group()
@opt_url
@click.pass_context
def cli(ctx, icepap):
    """
    High level command line interface for IcePAP
    """
    ctx.ensure_object(dict)
    ctx.obj["icepap"] = icepap


@cli.command()
@click.argument("pairs", nargs=-1, type=str, required=True)
@opt_pb_format
@opt_toolbar
@opt_title
@click.pass_context
def move(ctx, pairs, pb_format, bottom_toolbar, title):
    """
    Move specified axes in pairs <axis, position> to the specified absolute
    positions.
    """
    pap = ctx.obj["icepap"]
    motors = [pap[int(address)] for address in pairs[::2]]
    positions = [int(position) for position in pairs[1::2]]
    cli_move(Group(motors), positions, pb_format, bottom_toolbar, title)


@cli.command()
@click.argument("pairs", nargs=-1, type=str, required=True)
@opt_pb_format
@opt_toolbar
@opt_title
@click.option("-m", "--multiple", is_flag=True)
@click.pass_context
def rmove(ctx, pairs, pb_format, bottom_toolbar, title, multiple):
    """
    Move specified axes in pairs <axis, position> relative to their current
    positions.
    """
    pap = ctx.obj["icepap"]
    motors = [pap[int(address)] for address in pairs[::2]]
    deltas = [int(position) for position in pairs[1::2]]
    cli_rmove(
        Group(motors),
        deltas,
        pb_format,
        bottom_toolbar,
        title,
        multiple)


@cli.command()
@opt_mandatory_axes
def stop(motors):
    """Stops the given motors"""
    group = Group(motors)
    print_formatted_text(
        HTML("<orange>Stopping... </orange>"),
        end="",
        flush=True)
    group.start_stop()
    group.wait_stopped()
    print_formatted_text(HTML("[<green>DONE</green>]"))


@cli.command()
@opt_axes
@opt_table_style
def state(motors, table_style):
    """Prints a summary of each axis state in form of table"""
    group = Group(motors)
    click.echo(StateTable(group, style=table_style))


@cli.command()
@opt_axes
@opt_table_style
def status(motors, table_style):
    """
    Prints a summary of each axis status (position velotiy, acc. time, etc)
    in form of table
    """
    group = Group(motors)
    click.echo(StatusTable(group, style=table_style))


@cli.command()
@opt_axes
@opt_table_style
def pos(motors, table_style):
    """Prints a summary of each axis detailed position in form of table"""
    group = Group(motors)
    click.echo(PositionTable(group, style=table_style))


@cli.command()
@click.pass_context
def ver(ctx):
    """Prints a summary of icepap version"""
    pap = ctx.obj["icepap"]
    click.echo(pap.ver)


@cli.command()
@click.pass_context
def mode(ctx):
    """Prints the operation mode"""
    pap = ctx.obj["icepap"]
    click.echo(pap.mode)


@cli.command()
@opt_rack
@click.confirmation_option(
    prompt="Are you sure you want to reset the rack(s)?"
)
@click.pass_context
def reset(ctx, rack):
    """Resets the given rack. If no rack is given, resets the whole system"""
    pap = ctx.obj["icepap"]
    pap.reset(rack)


@cli.command()
@click.confirmation_option(
    prompt="Are you sure you want to reboot?"
)
@click.pass_context
def reboot(ctx):
    "Reboots the icepap"
    pap = ctx.obj["icepap"]
    pap.reboot()


@cli.command()
@opt_racks
@opt_table_style
@click.pass_context
def rinfo(ctx, racks, table_style):
    """Prints a summary of each rack information in form of table"""
    pap = ctx.obj["icepap"]
    rids = pap.get_rid(racks)
    temps = pap.get_rtemp(racks)
    table = Table(style=table_style)
    table.columns.header = ("Rack #", "RID", "Temp.")
    for row in zip(racks, rids, temps):
        table.rows.append(row)
    click.echo(table)


if __name__ == "__main__":
    cli()
